@@ -19,7 +19,8 @@ from datetime import datetime, timedelta
 
 from .exceptions import Forbidden, GeneralException, InternalError, InvalidData, NotFound
 from .response import Codes
-from .util import quote, hash, tzutc
+from .util import quote, hash, safeformat, now, absdatetime
+from . import __version__
 
 
 class Vingd:
@@ -33,6 +34,8 @@ class Vingd:
     EXP_ORDER = {'minutes': 15}
     EXP_VOUCHER = {'days': 7}
     
+    USER_AGENT = 'vingd-api-python/'+__version__
+    
     api_key = None
     api_secret = None
     api_endpoint = URL_ENDPOINT
@@ -66,7 +69,10 @@ class Vingd:
         port = 443
         path = urljoin(endpoint.path+'/', subpath)
         
-        headers = {'Authorization': 'Basic ' + base64.b64encode("%s:%s" % (self.api_key, self.api_secret)) }
+        headers = {
+            'Authorization': 'Basic ' + base64.b64encode("%s:%s" % (self.api_key, self.api_secret)),
+            'User-Agent': self.USER_AGENT
+        }
         try:
             conn = httplib.HTTPSConnection(host, port)
             conn.request(verb.upper(), quote(path), data, headers)
@@ -109,8 +115,8 @@ class Vingd:
     
     @staticmethod
     def _extract_id_from_batch_response(r, name='id'):
-        """ Unholy, forward-compatible, mess for extraction of id/oid from a
-        soon-to-be (deprecated) batch response. """
+        """Unholy, forward-compatible, mess for extraction of id/oid from a
+        soon-to-be (deprecated) batch response."""
         names = name + 's'
         if names in r:
             # soon-to-be deprecated batch reponse
@@ -121,6 +127,24 @@ class Vingd:
             # new-style simplified api response
             id = r[name]
         return int(id)
+
+    def kvpath(self, base, *pa, **kw):
+        """Key-value query url builder (of the form: "base/v0/k1=v1/k2=v2"). """
+        def fmt(vtuple):
+            typ, val = vtuple
+            if not val is None:
+                return safeformat("{:%s}"%typ, val)
+            return None
+        parts = [base]
+        for v in pa:
+            fv = fmt(v)
+            if not fv is None:
+                parts.append(quote(fv))
+        for k,v in kw.iteritems():
+            fv = fmt(v)
+            if not fv is None:
+                parts.append(quote(k+"="+fv))
+        return "/".join(parts)
     
     def create_object(self, name, url):
         """
@@ -200,7 +224,10 @@ class Vingd:
         :resource: ``objects/<oid>/tokens/<tid>``
         :access: authenticated user MUST be the object's owner
         """
-        return self.request('get', 'objects/%d/tokens/%s' % (oid, tid))
+        return self.request(
+            'get',
+            safeformat('objects/{:int}/tokens/{:hex}', oid, tid)
+        )
     
     def commit_purchase(self, purchaseid, transferid):
         """
@@ -234,7 +261,7 @@ class Vingd:
         """
         return self.request(
             'put',
-            'purchases/%d' % int(purchaseid),
+            safeformat('purchases/{:int}', purchaseid),
             json.dumps({'transferid': transferid})
         )
     
@@ -284,19 +311,19 @@ class Vingd:
         :resource: ``objects/<oid>/orders/``
         :access: authorized users
         """
-        if expires is None:
-            expires = self.EXP_ORDER
-        if isinstance(expires, dict):
-            expires = datetime.now(tzutc())+timedelta(**expires)
-        orders = self.request('post', 'objects/%d/orders/' % int(oid), json.dumps({
-            'price': price,
-            'order_expires': expires.isoformat(),
-            'context': context
-        }))
+        expires = absdatetime(expires, default=self.EXP_ORDER)
+        orders = self.request(
+            'post',
+            safeformat('objects/{:int}/orders/', oid),
+            json.dumps({
+                'price': price,
+                'order_expires': expires.isoformat(),
+                'context': context
+            }))
         orderid = self._extract_id_from_batch_response(orders)
         return {
             'id': orderid,
-            'expires': expires.isoformat(),
+            'expires': expires,
             'context': context,
             'object': {
                 'id': oid,
@@ -335,9 +362,9 @@ class Vingd:
         return self.request(
             'get',
             '%sorders/%s%s' % (
-                ('objects/%d/' % oid) if oid else "",
+                safeformat('objects/{:int}/', oid) if oid else "",
                 "all/" if include_expired else "",
-                orderid if orderid else ""
+                safeformat('{:int}', orderid) if orderid else ""
             )
         )
     
@@ -384,7 +411,8 @@ class Vingd:
         :access: authorized user MUST be the object owner
         """
         r = self.request(
-            'put', 'registry/objects/%d/' % oid,
+            'put',
+            safeformat('registry/objects/{:int}/', oid),
             json.dumps({
                 'description': {
                     'name': name,
@@ -403,14 +431,16 @@ class Vingd:
         :type oid: ``bigint``
         :param oid:
             Object ID
-        :type since: ``string``
+        :type since: ``datetime``/``dict``
         :param since:
-            Object has to be newer than this timestamp (in ISO 8601 basic
-            format).
-        :type until: ``string``
+            Object has to be newer than this timestamp (absolute ``datetime``,
+            or relative ``dict``). Valid keys for relative `since` timestamp
+            dictionary are same as keyword arguments for `datetime.timedelta`
+            (``days``, ``seconds``, ``minutes``, ``hours``, ``weeks``).
+        :type until: ``datetime``/``dict``
         :param until:
-            Object has to be older than this timestamp (in ISO 8601 basic
-            format).
+            Object has to be older than this timestamp (for format, see the
+            `since` parameter above).
         :type last: ``bigint``
         :param last:
             The number of newest objects (that satisfy all other criteria) to
@@ -432,12 +462,10 @@ class Vingd:
         :access: authorized users (only objects owned by the authenticated user
             are returned)
         """
-        resource = 'registry/objects'
-        if oid: resource += '/%d' % int(oid)
-        if since: resource += '/since=%s' % since
-        if until: resource += '/until=%s' % until
-        if first: resource += '/first=%d' % int(first)
-        if last: resource += '/last=%d' % int(last)
+        resource = self.kvpath('registry/objects', ('int', oid),
+                               since=('isobasic', absdatetime(since)),
+                               until=('isobasic', absdatetime(until)),
+                               first=('int', first), last=('int', last))
         return self.request('get', resource)
     
     def get_object(self, oid):
@@ -462,7 +490,7 @@ class Vingd:
         :access: authorized users (only objects owned by the authenticated user
             are returned)
         """
-        return self.request('get', 'registry/objects/%d' % oid)
+        return self.request('get', safeformat('registry/objects/{:int}', oid))
     
     def get_user_profile(self):
         """
@@ -505,7 +533,8 @@ class Vingd:
         :access: authorized users; delegate permission required for the
             requester to read user's balance: ``get.account.balance``
         """
-        return int(self.request('get', 'fort/accounts/%s' % huid)['balance'])
+        acc = self.request('get', safeformat('fort/accounts/{:hex}', huid))
+        return int(acc['balance'])
     
     def authorized_purchase_object(self, oid, price, huid):
         """Does delegated (pre-authorized) purchase of `oid` in the name of
@@ -519,11 +548,14 @@ class Vingd:
             delegate permission required for the requester to charge the
             user: ``purchase.object``
         """
-        return self.request('post', 'objects/%d/purchases' % oid, json.dumps({
-            'price': price,
-            'huid': huid,
-            'autocommit': True
-        }))
+        return self.request(
+            'post',
+            safeformat('objects/{:int}/purchases', oid),
+            json.dumps({
+                'price': price,
+                'huid': huid,
+                'autocommit': True
+            }))
     
     def authorized_create_user(self, identities, primary, permissions=None):
         """Creates Vingd user (profile & account), links it with the provided
@@ -635,13 +667,10 @@ class Vingd:
         :resource: ``vouchers/``
         :access: authorized users (ACL flag: ``voucher.add``)
         """
-        if expires is None:
-            expires = self.EXP_VOUCHER
-        if isinstance(expires, dict):
-            expires = datetime.now(tzutc())+timedelta(**expires)
+        expires = absdatetime(expires, default=self.EXP_VOUCHER).isoformat()
         voucher = self.request('post', 'vouchers/', json.dumps({
             'amount': amount,
-            'until': expires.isoformat(),
+            'until': expires,
             'message': message,
             'gid': gid
         }))
@@ -672,13 +701,17 @@ class Vingd:
         :type gid: ``alphanumeric(32)``
         :param gid:
             Filter by voucher Group ID. GID is localized to `uid_from`.
-        :type valid_after: ``string``
+        :type valid_after: ``datetime``/``dict``
         :param valid_after:
-            Voucher has to be valid after this timestamp (in ISO 8601 basic
-            format).
-        :type valid_before: ``string``
+            Voucher has to be valid after this timestamp. Absolute
+            (``datetime``) or relative (``dict``) timestamps are accepted. Valid
+            keys for relative timestamp dictionary are same as keyword arguments
+            for `datetime.timedelta` (``days``, ``seconds``, ``minutes``,
+            ``hours``, ``weeks``).
+        :type valid_before: ``datetime``/``dict``
         :param valid_before:
-            Voucher was valid until this timestamp (in ISO 8601 basic format).
+            Voucher was valid until this timestamp (for format, see the
+            `valid_after` above).
         :type last: ``bigint``
         :param last:
             The number of newest vouchers (that satisfy all other criteria) to
@@ -704,15 +737,19 @@ class Vingd:
             ``[/last=<last>][/first=<first>]``
         :access: authorized users (ACL flag: ``voucher.get``)
         """
-        resource = 'vouchers'
-        if vid_encoded: resource += '/%s' % vid_encoded
-        if uid_from: resource += '/from=%d' % int(uid_from)
-        if uid_to: resource += '/to=%d' % int(uid_to)
-        if gid: resource += '/gid=%s' % gid
-        if valid_after: resource += '/valid_after=%s' % valid_after
-        if valid_before: resource += '/valid_before=%s' % valid_before
-        if first: resource += '/first=%d' % int(first)
-        if last: resource += '/last=%d' % int(last)
+        resource = self.kvpath(
+            'vouchers',
+            ('ident', vid_encoded),
+            **{
+                'from': ('int', uid_from),
+                'to': ('int', uid_to),
+                'gid': ('ident', gid),
+                'valid_after': ('isobasic', absdatetime(valid_after)),
+                'valid_before': ('isobasic', absdatetime(valid_before)),
+                'first': ('int', first),
+                'last': ('int', last)
+            }
+        )
         return self.request('get', resource)
     
     def get_vouchers_history(self, vid_encoded=None, vid=None, action=None,
@@ -741,20 +778,25 @@ class Vingd:
         :type gid: ``alphanumeric(32)``
         :param gid:
             Filter by voucher Group ID. GID is localized to `uid_from`.
-        :type valid_after: ``string``
+        :type valid_after: ``datetime``/``dict``
         :param valid_after:
-            Voucher has to be valid after this timestamp (in ISO 8601 basic
-            format).
-        :type valid_before: ``string``
+            Voucher has to be valid after this timestamp. Absolute
+            (``datetime``) or relative (``dict``) timestamps are accepted. Valid
+            keys for relative timestamp dictionary are same as keyword arguments
+            for `datetime.timedelta` (``days``, ``seconds``, ``minutes``,
+            ``hours``, ``weeks``).
+        :type valid_before: ``datetime``/``dict``
         :param valid_before:
-            Voucher was valid until this timestamp (in ISO 8601 basic format).
-        :type create_after: ``string``
+            Voucher was valid until this timestamp (for format, see the
+            `valid_after` above).
+        :type create_after: ``datetime``/``dict``
         :param create_after:
-            Voucher has to be created after this timestamp (in ISO 8601 basic
-            format).
-        :type create_before: ``string``
+            Voucher has to be created after this timestamp (for format, see the
+            `valid_after` above).
+        :type create_before: ``datetime``/``dict``
         :param create_before:
-            Voucher was created until this timestamp (in ISO 8601 basic format).
+            Voucher was created until this timestamp (for format, see the
+            `valid_after` above).
         :type last: ``bigint``
         :param last:
             The number of newest voucher entries (that satisfy all other
@@ -781,19 +823,23 @@ class Vingd:
             ``[/gid=<group_id>]``
         :access: authorized users (ACL flag: ``voucher.history``)
         """
-        resource = 'vouchers/history'
-        if vid_encoded: resource += '/%s' % vid_encoded
-        if vid: resource += '/vid=%d' % long(vid)
-        if action: resource += '/action=%s' % action
-        if uid_from: resource += '/from=%d' % int(uid_from)
-        if uid_to: resource += '/to=%d' % int(uid_to)
-        if gid: resource += '/gid=%s' % gid
-        if valid_after: resource += '/valid_after=%s' % valid_after
-        if valid_before: resource += '/valid_before=%s' % valid_before
-        if create_after: resource += '/create_after=%s' % create_after
-        if create_before: resource += '/create_before=%s' % create_before
-        if first: resource += '/first=%d' % int(first)
-        if last: resource += '/last=%d' % int(last)
+        resource = self.kvpath(
+            'vouchers/history',
+            ('ident', vid_encoded),
+            **{
+                'vid': ('int', vid),
+                'action': ('ident', action),
+                'from': ('int', uid_from),
+                'to': ('int', uid_to),
+                'gid': ('ident', gid),
+                'valid_after': ('isobasic', absdatetime(valid_after)),
+                'valid_before': ('isobasic', absdatetime(valid_before)),
+                'create_after': ('isobasic', absdatetime(create_after)),
+                'create_before': ('isobasic', absdatetime(create_before)),
+                'first': ('int', first),
+                'last': ('int', last)
+            }
+        )
         return self.request('get', resource)
     
     def revoke_vouchers(self, vid_encoded=None,
@@ -815,13 +861,17 @@ class Vingd:
         :type gid: ``alphanumeric(32)``
         :param gid:
             Filter by voucher Group ID. GID is localized to `uid_from`.
-        :type valid_after: ``string``
+        :type valid_after: ``datetime``/``dict``
         :param valid_after:
-            Voucher has to be valid after this timestamp (in ISO 8601 basic
-            format).
-        :type valid_before: ``string``
+            Voucher has to be valid after this timestamp. Absolute
+            (``datetime``) or relative (``dict``) timestamps are accepted. Valid
+            keys for relative timestamp dictionary are same as keyword arguments
+            for `datetime.timedelta` (``days``, ``seconds``, ``minutes``,
+            ``hours``, ``weeks``).
+        :type valid_before: ``datetime``/``dict``
         :param valid_before:
-            Voucher was valid until this timestamp (in ISO 8601 basic format).
+            Voucher was valid until this timestamp (for format, see the
+            `valid_after` above).
         :type last: ``bigint``
         :param last:
             The number of newest vouchers (that satisfy all other criteria) to
@@ -851,13 +901,17 @@ class Vingd:
             ``[/last=<last>][/first=<first>]``
         :access: authorized users (ACL flag: ``voucher.revoke``)
         """
-        resource = 'vouchers'
-        if vid_encoded: resource += '/%s' % vid_encoded
-        if uid_from: resource += '/from=%d' % int(uid_from)
-        if uid_to: resource += '/to=%d' % int(uid_to)
-        if gid: resource += '/gid=%s' % gid
-        if valid_after: resource += '/valid_after=%s' % valid_after
-        if valid_before: resource += '/valid_before=%s' % valid_before
-        if first: resource += '/first=%d' % int(first)
-        if last: resource += '/last=%d' % int(last)
+        resource = self.kvpath(
+            'vouchers',
+            ('ident', vid_encoded),
+            **{
+                'from': ('int', uid_from),
+                'to': ('int', uid_to),
+                'gid': ('ident', gid),
+                'valid_after': ('isobasic', absdatetime(valid_after)),
+                'valid_before': ('isobasic', absdatetime(valid_before)),
+                'first': ('int', first),
+                'last': ('int', last)
+            }
+        )
         return self.request('delete', resource, json.dumps({'revoke': True}))
